/*
 * @BEGIN LICENSE
 *
 * Psi4: an open-source quantum chemistry software package
 *
 * Copyright (c) 2007-2017 The Psi4 Developers.
 *
 * The copyrights for code used from other parties are included in
 * the corresponding files.
 *
 * This file is part of Psi4.
 *
 * Psi4 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * Psi4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with Psi4; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * @END LICENSE
 */

/*
** The new PSI standard header file for atomic masses.
**
** Created by Matt Leininger, July 1995
**
** This file contains three arrays.  The first is a list of atomic masses
** for a number of isotopes.  The second is a list of isotope labels which
** correspond to the masses in the first array.  The mass labels given
** are the same as those in W. D. Allen's INTDER95.  The final array is a 
** list of atomic masses for the first several elements, with the most
** common isotopomer mass given.  This last array is most useful for
** converting an atomic number into an atomic mass. 
**
** Please be sure that any modifications to this file are reflected in
** _all three_ arrays if necessary.
**
** Modifications:
** 10/08/99 EFV - Changed an2masses[0] to 0.0000 
**                (ghost atom's weight should be 0)
**
** 08/16/02 EFV - Updated atomic_masses[] and an2masses for H-Ar using NIST database
**                of Atomic Weights and Isotopic Compositions (www.nist.gov)
**
** 08/23/02 EFV - Use lithium isotope 7 as the default (most abundant) 
**
** 03/03/08 MT - extended set of atoms by data from mathematica script. For an2masses
**               for naturally occurring atoms, the most abundant is used, otherwise the
**               atom with the longest halflife time.
**
*/

#ifndef _psi_include_masses_h_
#define _psi_include_masses_h_


#define LAST_ATOMIC_INDEX (118)

static const char *atomic_labels[]= 
{"X","H","HE","LI","BE","B","C","N","O","F","NE","NA","MG","AL","SI",
"P","S","CL","AR","K","CA","SC","TI","V","CR","MN","FE","CO","NI",
"CU","ZN","GA","GE","AS","SE","BR","KR","RB","SR","Y","ZR","NB","MO",
"TC","RU","RH","PD","AG","CD","IN","SN","SB","TE","I","XE","CS","BA",
"LA","CE","PR","ND","PM","SM","EU","GD","TB","DY","HO","ER","TM","YB",
"LU","HF","TA","W","RE","OS","IR","PT","AU","HG","TL","PB","BI","PO",
"AT","RN","FR","RA","AC","TH","PA","U","NP","PU","AM","CM","BK","CF",
"ES","FM","MD","NO","LR","RF","DB","SG","BH","HS","MT","DS","RG",
  "UUB","UUT","UUQ","UUP","UUH","UUS","UUO"};

static const double an2masses[]=
{0.,1.00782503207,4.00260325415,7.016004548,9.012182201,11.009305406,
12,14.00307400478,15.99491461956,18.998403224,19.99244017542,
22.98976928087,23.985041699,26.981538627,27.97692653246,30.973761629,
31.972070999,34.968852682,39.96238312251,38.963706679,39.962590983,
44.955911909,47.947946281,50.943959507,51.940507472,54.938045141,
55.934937475,58.933195048,57.935342907,62.929597474,63.929142222,
68.925573587,73.921177767,74.921596478,79.916521271,78.918337087,
85.910610729,84.911789737,87.905612124,88.905848295,89.904704416,
92.906378058,97.905408169,98.906254747,101.904349312,102.905504292,
105.903485715,106.90509682,113.90335854,114.903878484,119.902194676,
120.903815686,129.906224399,126.904472681,131.904153457,132.905451932,
137.905247237,138.906353267,139.905438706,140.907652769,141.907723297,
144.912749023,151.919732425,152.921230339,157.924103912,158.925346757,
163.929174751,164.93032207,165.930293061,168.93421325,173.938862089,
174.940771819,179.946549953,180.947995763,183.950931188,186.955753109,
191.96148069,192.96292643,194.964791134,196.966568662,201.970643011,
204.974427541,207.976652071,208.980398734,208.982430435,210.987496271,
222.017577738,222.01755173,228.031070292,227.027752127,232.038055325,
231.03588399,238.050788247,237.048173444,242.058742611,243.06138108,
247.07035354,247.07030708,251.079586788,252.082978512,257.095104724,
258.098431319,255.093241131,260.105504,263.112547,255.107398,259.114500,
262.122892,263.128558,265.136151,281.162061,272.153615,283.171792,283.176451,
285.183698,287.191186,292.199786,291.206564,293.214670};

#define LAST_MASS_INDEX (3297)

static const char *mass_labels[]=
{"H","H1","H2","D","H3","T","H4","H5","H6","H7","HE","HE3","HE4",
"HE5","HE6","HE7","HE8","HE9","HE10","LI","LI3","LI4","LI5","LI6",
"LI7","LI8","LI9","LI10","LI11","LI12","BE","BE5","BE6","BE7","BE8",
"BE9","BE10","BE11","BE12","BE13","BE14","BE15","BE16","B","B6","B7",
"B8","B9","B10","B11","B12","B13","B14","B15","B16","B17","B18","B19",
"C","C8","C9","C10","C11","C12","C13","C14","C15","C16","C17","C18",
"C19","C20","C21","C22","N","N10","N11","N12","N13","N14","N15","N16",
"N17","N18","N19","N20","N21","N22","N23","N24","N25","O","O12","O13",
"O14","O15","O16","O17","O18","O19","O20","O21","O22","O23","O24",
"O25","O26","O27","O28","F","F14","F15","F16","F17","F18","F19","F20",
"F21","F22","F23","F24","F25","F26","F27","F28","F29","F30","F31",
"NE","NE16","NE17","NE18","NE19","NE20","NE21","NE22","NE23","NE24",
"NE25","NE26","NE27","NE28","NE29","NE30","NE31","NE32","NE33","NE34",
"NA","NA18","NA19","NA20","NA21","NA22","NA23","NA24","NA25","NA26",
"NA27","NA28","NA29","NA30","NA31","NA32","NA33","NA34","NA35","NA36",
"NA37","MG","MG19","MG20","MG21","MG22","MG23","MG24","MG25","MG26",
"MG27","MG28","MG29","MG30","MG31","MG32","MG33","MG34","MG35","MG36",
"MG37","MG38","MG39","MG40","AL","AL21","AL22","AL23","AL24","AL25",
"AL26","AL27","AL28","AL29","AL30","AL31","AL32","AL33","AL34","AL35",
"AL36","AL37","AL38","AL39","AL40","AL41","AL42","SI","SI22","SI23",
"SI24","SI25","SI26","SI27","SI28","SI29","SI30","SI31","SI32","SI33",
"SI34","SI35","SI36","SI37","SI38","SI39","SI40","SI41","SI42","SI43",
"SI44","P","P24","P25","P26","P27","P28","P29","P30","P31","P32",
"P33","P34","P35","P36","P37","P38","P39","P40","P41","P42","P43",
"P44","P45","P46","S","S26","S27","S28","S29","S30","S31","S32","S33",
"S34","S35","S36","S37","S38","S39","S40","S41","S42","S43","S44",
"S45","S46","S47","S48","S49","CL","CL28","CL29","CL30","CL31","CL32",
"CL33","CL34","CL35","CL36","CL37","CL38","CL39","CL40","CL41","CL42",
"CL43","CL44","CL45","CL46","CL47","CL48","CL49","CL50","CL51","AR",
"AR30","AR31","AR32","AR33","AR34","AR35","AR36","AR37","AR38","AR39",
"AR40","AR41","AR42","AR43","AR44","AR45","AR46","AR47","AR48","AR49",
"AR50","AR51","AR52","AR53","K","K32","K33","K34","K35","K36","K37",
"K38","K39","K40","K41","K42","K43","K44","K45","K46","K47","K48",
"K49","K50","K51","K52","K53","K54","K55","CA","CA34","CA35","CA36",
"CA37","CA38","CA39","CA40","CA41","CA42","CA43","CA44","CA45","CA46",
"CA47","CA48","CA49","CA50","CA51","CA52","CA53","CA54","CA55","CA56",
"CA57","SC","SC36","SC37","SC38","SC39","SC40","SC41","SC42","SC43",
"SC44","SC45","SC46","SC47","SC48","SC49","SC50","SC51","SC52","SC53",
"SC54","SC55","SC56","SC57","SC58","SC59","SC60","TI","TI38","TI39",
"TI40","TI41","TI42","TI43","TI44","TI45","TI46","TI47","TI48","TI49",
"TI50","TI51","TI52","TI53","TI54","TI55","TI56","TI57","TI58","TI59",
"TI60","TI61","TI62","TI63","V","V40","V41","V42","V43","V44","V45",
"V46","V47","V48","V49","V50","V51","V52","V53","V54","V55","V56",
"V57","V58","V59","V60","V61","V62","V63","V64","V65","CR","CR42",
"CR43","CR44","CR45","CR46","CR47","CR48","CR49","CR50","CR51","CR52",
"CR53","CR54","CR55","CR56","CR57","CR58","CR59","CR60","CR61","CR62",
"CR63","CR64","CR65","CR66","CR67","MN","MN44","MN45","MN46","MN47",
"MN48","MN49","MN50","MN51","MN52","MN53","MN54","MN55","MN56","MN57",
"MN58","MN59","MN60","MN61","MN62","MN63","MN64","MN65","MN66","MN67",
"MN68","MN69","FE","FE45","FE46","FE47","FE48","FE49","FE50","FE51",
"FE52","FE53","FE54","FE55","FE56","FE57","FE58","FE59","FE60","FE61",
"FE62","FE63","FE64","FE65","FE66","FE67","FE68","FE69","FE70","FE71",
"FE72","CO","CO47","CO48","CO49","CO50","CO51","CO52","CO53","CO54",
"CO55","CO56","CO57","CO58","CO59","CO60","CO61","CO62","CO63","CO64",
"CO65","CO66","CO67","CO68","CO69","CO70","CO71","CO72","CO73","CO74",
"CO75","NI","NI48","NI49","NI50","NI51","NI52","NI53","NI54","NI55",
"NI56","NI57","NI58","NI59","NI60","NI61","NI62","NI63","NI64","NI65",
"NI66","NI67","NI68","NI69","NI70","NI71","NI72","NI73","NI74","NI75",
"NI76","NI77","NI78","CU","CU52","CU53","CU54","CU55","CU56","CU57",
"CU58","CU59","CU60","CU61","CU62","CU63","CU64","CU65","CU66","CU67",
"CU68","CU69","CU70","CU71","CU72","CU73","CU74","CU75","CU76","CU77",
"CU78","CU79","CU80","ZN","ZN54","ZN55","ZN56","ZN57","ZN58","ZN59",
"ZN60","ZN61","ZN62","ZN63","ZN64","ZN65","ZN66","ZN67","ZN68","ZN69",
"ZN70","ZN71","ZN72","ZN73","ZN74","ZN75","ZN76","ZN77","ZN78","ZN79",
"ZN80","ZN81","ZN82","ZN83","GA","GA56","GA57","GA58","GA59","GA60",
"GA61","GA62","GA63","GA64","GA65","GA66","GA67","GA68","GA69","GA70",
"GA71","GA72","GA73","GA74","GA75","GA76","GA77","GA78","GA79","GA80",
"GA81","GA82","GA83","GA84","GA85","GA86","GE","GE58","GE59","GE60",
"GE61","GE62","GE63","GE64","GE65","GE66","GE67","GE68","GE69","GE70",
"GE71","GE72","GE73","GE74","GE75","GE76","GE77","GE78","GE79","GE80",
"GE81","GE82","GE83","GE84","GE85","GE86","GE87","GE88","GE89","AS",
"AS60","AS61","AS62","AS63","AS64","AS65","AS66","AS67","AS68","AS69",
"AS70","AS71","AS72","AS73","AS74","AS75","AS76","AS77","AS78","AS79",
"AS80","AS81","AS82","AS83","AS84","AS85","AS86","AS87","AS88","AS89",
"AS90","AS91","AS92","SE","SE65","SE66","SE67","SE68","SE69","SE70",
"SE71","SE72","SE73","SE74","SE75","SE76","SE77","SE78","SE79","SE80",
"SE81","SE82","SE83","SE84","SE85","SE86","SE87","SE88","SE89","SE90",
"SE91","SE92","SE93","SE94","BR","BR67","BR68","BR69","BR70","BR71",
"BR72","BR73","BR74","BR75","BR76","BR77","BR78","BR79","BR80","BR81",
"BR82","BR83","BR84","BR85","BR86","BR87","BR88","BR89","BR90","BR91",
"BR92","BR93","BR94","BR95","BR96","BR97","KR","KR69","KR70","KR71",
"KR72","KR73","KR74","KR75","KR76","KR77","KR78","KR79","KR80","KR81",
"KR82","KR83","KR84","KR85","KR86","KR87","KR88","KR89","KR90","KR91",
"KR92","KR93","KR94","KR95","KR96","KR97","KR98","KR99","KR100","RB",
"RB71","RB72","RB73","RB74","RB75","RB76","RB77","RB78","RB79","RB80",
"RB81","RB82","RB83","RB84","RB85","RB86","RB87","RB88","RB89","RB90",
"RB91","RB92","RB93","RB94","RB95","RB96","RB97","RB98","RB99",
"RB100","RB101","RB102","SR","SR73","SR74","SR75","SR76","SR77",
"SR78","SR79","SR80","SR81","SR82","SR83","SR84","SR85","SR86","SR87",
"SR88","SR89","SR90","SR91","SR92","SR93","SR94","SR95","SR96","SR97",
"SR98","SR99","SR100","SR101","SR102","SR103","SR104","SR105","Y",
"Y76","Y77","Y78","Y79","Y80","Y81","Y82","Y83","Y84","Y85","Y86",
"Y87","Y88","Y89","Y90","Y91","Y92","Y93","Y94","Y95","Y96","Y97",
"Y98","Y99","Y100","Y101","Y102","Y103","Y104","Y105","Y106","Y107",
"Y108","ZR","ZR78","ZR79","ZR80","ZR81","ZR82","ZR83","ZR84","ZR85",
"ZR86","ZR87","ZR88","ZR89","ZR90","ZR91","ZR92","ZR93","ZR94","ZR95",
"ZR96","ZR97","ZR98","ZR99","ZR100","ZR101","ZR102","ZR103","ZR104",
"ZR105","ZR106","ZR107","ZR108","ZR109","ZR110","NB","NB81","NB82",
"NB83","NB84","NB85","NB86","NB87","NB88","NB89","NB90","NB91","NB92",
"NB93","NB94","NB95","NB96","NB97","NB98","NB99","NB100","NB101",
"NB102","NB103","NB104","NB105","NB106","NB107","NB108","NB109",
"NB110","NB111","NB112","NB113","MO","MO83","MO84","MO85","MO86",
"MO87","MO88","MO89","MO90","MO91","MO92","MO93","MO94","MO95","MO96",
"MO97","MO98","MO99","MO100","MO101","MO102","MO103","MO104","MO105",
"MO106","MO107","MO108","MO109","MO110","MO111","MO112","MO113",
"MO114","MO115","TC","TC85","TC86","TC87","TC88","TC89","TC90","TC91",
"TC92","TC93","TC94","TC95","TC96","TC97","TC98","TC99","TC100",
"TC101","TC102","TC103","TC104","TC105","TC106","TC107","TC108",
"TC109","TC110","TC111","TC112","TC113","TC114","TC115","TC116",
"TC117","TC118","RU","RU87","RU88","RU89","RU90","RU91","RU92","RU93",
"RU94","RU95","RU96","RU97","RU98","RU99","RU100","RU101","RU102",
"RU103","RU104","RU105","RU106","RU107","RU108","RU109","RU110",
"RU111","RU112","RU113","RU114","RU115","RU116","RU117","RU118",
"RU119","RU120","RH","RH89","RH90","RH91","RH92","RH93","RH94","RH95",
"RH96","RH97","RH98","RH99","RH100","RH101","RH102","RH103","RH104",
"RH105","RH106","RH107","RH108","RH109","RH110","RH111","RH112",
"RH113","RH114","RH115","RH116","RH117","RH118","RH119","RH120",
"RH121","RH122","PD","PD91","PD92","PD93","PD94","PD95","PD96","PD97",
"PD98","PD99","PD100","PD101","PD102","PD103","PD104","PD105","PD106",
"PD107","PD108","PD109","PD110","PD111","PD112","PD113","PD114",
"PD115","PD116","PD117","PD118","PD119","PD120","PD121","PD122",
"PD123","PD124","AG","AG93","AG94","AG95","AG96","AG97","AG98","AG99",
"AG100","AG101","AG102","AG103","AG104","AG105","AG106","AG107",
"AG108","AG109","AG110","AG111","AG112","AG113","AG114","AG115",
"AG116","AG117","AG118","AG119","AG120","AG121","AG122","AG123",
"AG124","AG125","AG126","AG127","AG128","AG129","AG130","CD","CD95",
"CD96","CD97","CD98","CD99","CD100","CD101","CD102","CD103","CD104",
"CD105","CD106","CD107","CD108","CD109","CD110","CD111","CD112",
"CD113","CD114","CD115","CD116","CD117","CD118","CD119","CD120",
"CD121","CD122","CD123","CD124","CD125","CD126","CD127","CD128",
"CD129","CD130","CD131","CD132","IN","IN97","IN98","IN99","IN100",
"IN101","IN102","IN103","IN104","IN105","IN106","IN107","IN108",
"IN109","IN110","IN111","IN112","IN113","IN114","IN115","IN116",
"IN117","IN118","IN119","IN120","IN121","IN122","IN123","IN124",
"IN125","IN126","IN127","IN128","IN129","IN130","IN131","IN132",
"IN133","IN134","IN135","SN","SN99","SN100","SN101","SN102","SN103",
"SN104","SN105","SN106","SN107","SN108","SN109","SN110","SN111",
"SN112","SN113","SN114","SN115","SN116","SN117","SN118","SN119",
"SN120","SN121","SN122","SN123","SN124","SN125","SN126","SN127",
"SN128","SN129","SN130","SN131","SN132","SN133","SN134","SN135",
"SN136","SN137","SB","SB103","SB104","SB105","SB106","SB107","SB108",
"SB109","SB110","SB111","SB112","SB113","SB114","SB115","SB116",
"SB117","SB118","SB119","SB120","SB121","SB122","SB123","SB124",
"SB125","SB126","SB127","SB128","SB129","SB130","SB131","SB132",
"SB133","SB134","SB135","SB136","SB137","SB138","SB139","TE","TE105",
"TE106","TE107","TE108","TE109","TE110","TE111","TE112","TE113",
"TE114","TE115","TE116","TE117","TE118","TE119","TE120","TE121",
"TE122","TE123","TE124","TE125","TE126","TE127","TE128","TE129",
"TE130","TE131","TE132","TE133","TE134","TE135","TE136","TE137",
"TE138","TE139","TE140","TE141","TE142","I","I108","I109","I110",
"I111","I112","I113","I114","I115","I116","I117","I118","I119","I120",
"I121","I122","I123","I124","I125","I126","I127","I128","I129","I130",
"I131","I132","I133","I134","I135","I136","I137","I138","I139","I140",
"I141","I142","I143","I144","XE","XE110","XE111","XE112","XE113",
"XE114","XE115","XE116","XE117","XE118","XE119","XE120","XE121",
"XE122","XE123","XE124","XE125","XE126","XE127","XE128","XE129",
"XE130","XE131","XE132","XE133","XE134","XE135","XE136","XE137",
"XE138","XE139","XE140","XE141","XE142","XE143","XE144","XE145",
"XE146","XE147","CS","CS112","CS113","CS114","CS115","CS116","CS117",
"CS118","CS119","CS120","CS121","CS122","CS123","CS124","CS125",
"CS126","CS127","CS128","CS129","CS130","CS131","CS132","CS133",
"CS134","CS135","CS136","CS137","CS138","CS139","CS140","CS141",
"CS142","CS143","CS144","CS145","CS146","CS147","CS148","CS149",
"CS150","CS151","BA","BA114","BA115","BA116","BA117","BA118","BA119",
"BA120","BA121","BA122","BA123","BA124","BA125","BA126","BA127",
"BA128","BA129","BA130","BA131","BA132","BA133","BA134","BA135",
"BA136","BA137","BA138","BA139","BA140","BA141","BA142","BA143",
"BA144","BA145","BA146","BA147","BA148","BA149","BA150","BA151",
"BA152","BA153","LA","LA117","LA118","LA119","LA120","LA121","LA122",
"LA123","LA124","LA125","LA126","LA127","LA128","LA129","LA130",
"LA131","LA132","LA133","LA134","LA135","LA136","LA137","LA138",
"LA139","LA140","LA141","LA142","LA143","LA144","LA145","LA146",
"LA147","LA148","LA149","LA150","LA151","LA152","LA153","LA154",
"LA155","CE","CE119","CE120","CE121","CE122","CE123","CE124","CE125",
"CE126","CE127","CE128","CE129","CE130","CE131","CE132","CE133",
"CE134","CE135","CE136","CE137","CE138","CE139","CE140","CE141",
"CE142","CE143","CE144","CE145","CE146","CE147","CE148","CE149",
"CE150","CE151","CE152","CE153","CE154","CE155","CE156","CE157","PR",
"PR121","PR122","PR123","PR124","PR125","PR126","PR127","PR128",
"PR129","PR130","PR131","PR132","PR133","PR134","PR135","PR136",
"PR137","PR138","PR139","PR140","PR141","PR142","PR143","PR144",
"PR145","PR146","PR147","PR148","PR149","PR150","PR151","PR152",
"PR153","PR154","PR155","PR156","PR157","PR158","PR159","ND","ND124",
"ND125","ND126","ND127","ND128","ND129","ND130","ND131","ND132",
"ND133","ND134","ND135","ND136","ND137","ND138","ND139","ND140",
"ND141","ND142","ND143","ND144","ND145","ND146","ND147","ND148",
"ND149","ND150","ND151","ND152","ND153","ND154","ND155","ND156",
"ND157","ND158","ND159","ND160","ND161","PM","PM126","PM127","PM128",
"PM129","PM130","PM131","PM132","PM133","PM134","PM135","PM136",
"PM137","PM138","PM139","PM140","PM141","PM142","PM143","PM144",
"PM145","PM146","PM147","PM148","PM149","PM150","PM151","PM152",
"PM153","PM154","PM155","PM156","PM157","PM158","PM159","PM160",
"PM161","PM162","PM163","SM","SM128","SM129","SM130","SM131","SM132",
"SM133","SM134","SM135","SM136","SM137","SM138","SM139","SM140",
"SM141","SM142","SM143","SM144","SM145","SM146","SM147","SM148",
"SM149","SM150","SM151","SM152","SM153","SM154","SM155","SM156",
"SM157","SM158","SM159","SM160","SM161","SM162","SM163","SM164",
"SM165","EU","EU130","EU131","EU132","EU133","EU134","EU135","EU136",
"EU137","EU138","EU139","EU140","EU141","EU142","EU143","EU144",
"EU145","EU146","EU147","EU148","EU149","EU150","EU151","EU152",
"EU153","EU154","EU155","EU156","EU157","EU158","EU159","EU160",
"EU161","EU162","EU163","EU164","EU165","EU166","EU167","GD","GD134",
"GD135","GD136","GD137","GD138","GD139","GD140","GD141","GD142",
"GD143","GD144","GD145","GD146","GD147","GD148","GD149","GD150",
"GD151","GD152","GD153","GD154","GD155","GD156","GD157","GD158",
"GD159","GD160","GD161","GD162","GD163","GD164","GD165","GD166",
"GD167","GD168","GD169","TB","TB136","TB137","TB138","TB139","TB140",
"TB141","TB142","TB143","TB144","TB145","TB146","TB147","TB148",
"TB149","TB150","TB151","TB152","TB153","TB154","TB155","TB156",
"TB157","TB158","TB159","TB160","TB161","TB162","TB163","TB164",
"TB165","TB166","TB167","TB168","TB169","TB170","TB171","DY","DY138",
"DY139","DY140","DY141","DY142","DY143","DY144","DY145","DY146",
"DY147","DY148","DY149","DY150","DY151","DY152","DY153","DY154",
"DY155","DY156","DY157","DY158","DY159","DY160","DY161","DY162",
"DY163","DY164","DY165","DY166","DY167","DY168","DY169","DY170",
"DY171","DY172","DY173","HO","HO140","HO141","HO142","HO143","HO144",
"HO145","HO146","HO147","HO148","HO149","HO150","HO151","HO152",
"HO153","HO154","HO155","HO156","HO157","HO158","HO159","HO160",
"HO161","HO162","HO163","HO164","HO165","HO166","HO167","HO168",
"HO169","HO170","HO171","HO172","HO173","HO174","HO175","ER","ER143",
"ER144","ER145","ER146","ER147","ER148","ER149","ER150","ER151",
"ER152","ER153","ER154","ER155","ER156","ER157","ER158","ER159",
"ER160","ER161","ER162","ER163","ER164","ER165","ER166","ER167",
"ER168","ER169","ER170","ER171","ER172","ER173","ER174","ER175",
"ER176","ER177","TM","TM145","TM146","TM147","TM148","TM149","TM150",
"TM151","TM152","TM153","TM154","TM155","TM156","TM157","TM158",
"TM159","TM160","TM161","TM162","TM163","TM164","TM165","TM166",
"TM167","TM168","TM169","TM170","TM171","TM172","TM173","TM174",
"TM175","TM176","TM177","TM178","TM179","YB","YB148","YB149","YB150",
"YB151","YB152","YB153","YB154","YB155","YB156","YB157","YB158",
"YB159","YB160","YB161","YB162","YB163","YB164","YB165","YB166",
"YB167","YB168","YB169","YB170","YB171","YB172","YB173","YB174",
"YB175","YB176","YB177","YB178","YB179","YB180","YB181","LU","LU150",
"LU151","LU152","LU153","LU154","LU155","LU156","LU157","LU158",
"LU159","LU160","LU161","LU162","LU163","LU164","LU165","LU166",
"LU167","LU168","LU169","LU170","LU171","LU172","LU173","LU174",
"LU175","LU176","LU177","LU178","LU179","LU180","LU181","LU182",
"LU183","LU184","HF","HF153","HF154","HF155","HF156","HF157","HF158",
"HF159","HF160","HF161","HF162","HF163","HF164","HF165","HF166",
"HF167","HF168","HF169","HF170","HF171","HF172","HF173","HF174",
"HF175","HF176","HF177","HF178","HF179","HF180","HF181","HF182",
"HF183","HF184","HF185","HF186","HF187","HF188","TA","TA155","TA156",
"TA157","TA158","TA159","TA160","TA161","TA162","TA163","TA164",
"TA165","TA166","TA167","TA168","TA169","TA170","TA171","TA172",
"TA173","TA174","TA175","TA176","TA177","TA178","TA179","TA180",
"TA181","TA182","TA183","TA184","TA185","TA186","TA187","TA188",
"TA189","TA190","W","W158","W159","W160","W161","W162","W163","W164",
"W165","W166","W167","W168","W169","W170","W171","W172","W173","W174",
"W175","W176","W177","W178","W179","W180","W181","W182","W183","W184",
"W185","W186","W187","W188","W189","W190","W191","W192","RE","RE160",
"RE161","RE162","RE163","RE164","RE165","RE166","RE167","RE168",
"RE169","RE170","RE171","RE172","RE173","RE174","RE175","RE176",
"RE177","RE178","RE179","RE180","RE181","RE182","RE183","RE184",
"RE185","RE186","RE187","RE188","RE189","RE190","RE191","RE192",
"RE193","RE194","OS","OS162","OS163","OS164","OS165","OS166","OS167",
"OS168","OS169","OS170","OS171","OS172","OS173","OS174","OS175",
"OS176","OS177","OS178","OS179","OS180","OS181","OS182","OS183",
"OS184","OS185","OS186","OS187","OS188","OS189","OS190","OS191",
"OS192","OS193","OS194","OS195","OS196","IR","IR164","IR165","IR166",
"IR167","IR168","IR169","IR170","IR171","IR172","IR173","IR174",
"IR175","IR176","IR177","IR178","IR179","IR180","IR181","IR182",
"IR183","IR184","IR185","IR186","IR187","IR188","IR189","IR190",
"IR191","IR192","IR193","IR194","IR195","IR196","IR197","IR198",
"IR199","PT","PT166","PT167","PT168","PT169","PT170","PT171","PT172",
"PT173","PT174","PT175","PT176","PT177","PT178","PT179","PT180",
"PT181","PT182","PT183","PT184","PT185","PT186","PT187","PT188",
"PT189","PT190","PT191","PT192","PT193","PT194","PT195","PT196",
"PT197","PT198","PT199","PT200","PT201","PT202","AU","AU169","AU170",
"AU171","AU172","AU173","AU174","AU175","AU176","AU177","AU178",
"AU179","AU180","AU181","AU182","AU183","AU184","AU185","AU186",
"AU187","AU188","AU189","AU190","AU191","AU192","AU193","AU194",
"AU195","AU196","AU197","AU198","AU199","AU200","AU201","AU202",
"AU203","AU204","AU205","HG","HG171","HG172","HG173","HG174","HG175",
"HG176","HG177","HG178","HG179","HG180","HG181","HG182","HG183",
"HG184","HG185","HG186","HG187","HG188","HG189","HG190","HG191",
"HG192","HG193","HG194","HG195","HG196","HG197","HG198","HG199",
"HG200","HG201","HG202","HG203","HG204","HG205","HG206","HG207",
"HG208","HG209","HG210","TL","TL176","TL177","TL178","TL179","TL180",
"TL181","TL182","TL183","TL184","TL185","TL186","TL187","TL188",
"TL189","TL190","TL191","TL192","TL193","TL194","TL195","TL196",
"TL197","TL198","TL199","TL200","TL201","TL202","TL203","TL204",
"TL205","TL206","TL207","TL208","TL209","TL210","TL211","TL212","PB",
"PB178","PB179","PB180","PB181","PB182","PB183","PB184","PB185",
"PB186","PB187","PB188","PB189","PB190","PB191","PB192","PB193",
"PB194","PB195","PB196","PB197","PB198","PB199","PB200","PB201",
"PB202","PB203","PB204","PB205","PB206","PB207","PB208","PB209",
"PB210","PB211","PB212","PB213","PB214","PB215","BI","BI184","BI185",
"BI186","BI187","BI188","BI189","BI190","BI191","BI192","BI193",
"BI194","BI195","BI196","BI197","BI198","BI199","BI200","BI201",
"BI202","BI203","BI204","BI205","BI206","BI207","BI208","BI209",
"BI210","BI211","BI212","BI213","BI214","BI215","BI216","BI217",
"BI218","PO","PO188","PO189","PO190","PO191","PO192","PO193","PO194",
"PO195","PO196","PO197","PO198","PO199","PO200","PO201","PO202",
"PO203","PO204","PO205","PO206","PO207","PO208","PO209","PO210",
"PO211","PO212","PO213","PO214","PO215","PO216","PO217","PO218",
"PO219","PO220","AT","AT193","AT194","AT195","AT196","AT197","AT198",
"AT199","AT200","AT201","AT202","AT203","AT204","AT205","AT206",
"AT207","AT208","AT209","AT210","AT211","AT212","AT213","AT214",
"AT215","AT216","AT217","AT218","AT219","AT220","AT221","AT222",
"AT223","RN","RN195","RN196","RN197","RN198","RN199","RN200","RN201",
"RN202","RN203","RN204","RN205","RN206","RN207","RN208","RN209",
"RN210","RN211","RN212","RN213","RN214","RN215","RN216","RN217",
"RN218","RN219","RN220","RN221","RN222","RN223","RN224","RN225",
"RN226","RN227","RN228","FR","FR199","FR200","FR201","FR202","FR203",
"FR204","FR205","FR206","FR207","FR208","FR209","FR210","FR211",
"FR212","FR213","FR214","FR215","FR216","FR217","FR218","FR219",
"FR220","FR221","FR222","FR223","FR224","FR225","FR226","FR227",
"FR228","FR229","FR230","FR231","FR232","RA","RA202","RA203","RA204",
"RA205","RA206","RA207","RA208","RA209","RA210","RA211","RA212",
"RA213","RA214","RA215","RA216","RA217","RA218","RA219","RA220",
"RA221","RA222","RA223","RA224","RA225","RA226","RA227","RA228",
"RA229","RA230","RA231","RA232","RA233","RA234","AC","AC206","AC207",
"AC208","AC209","AC210","AC211","AC212","AC213","AC214","AC215",
"AC216","AC217","AC218","AC219","AC220","AC221","AC222","AC223",
"AC224","AC225","AC226","AC227","AC228","AC229","AC230","AC231",
"AC232","AC233","AC234","AC235","AC236","TH","TH209","TH210","TH211",
"TH212","TH213","TH214","TH215","TH216","TH217","TH218","TH219",
"TH220","TH221","TH222","TH223","TH224","TH225","TH226","TH227",
"TH228","TH229","TH230","TH231","TH232","TH233","TH234","TH235",
"TH236","TH237","TH238","PA","PA212","PA213","PA214","PA215","PA216",
"PA217","PA218","PA219","PA220","PA221","PA222","PA223","PA224",
"PA225","PA226","PA227","PA228","PA229","PA230","PA231","PA232",
"PA233","PA234","PA235","PA236","PA237","PA238","PA239","PA240","U",
"U217","U218","U219","U220","U221","U222","U223","U224","U225","U226",
"U227","U228","U229","U230","U231","U232","U233","U234","U235","U236",
"U237","U238","U239","U240","U241","U242","NP","NP225","NP226",
"NP227","NP228","NP229","NP230","NP231","NP232","NP233","NP234",
"NP235","NP236","NP237","NP238","NP239","NP240","NP241","NP242",
"NP243","NP244","PU","PU228","PU229","PU230","PU231","PU232","PU233",
"PU234","PU235","PU236","PU237","PU238","PU239","PU240","PU241",
"PU242","PU243","PU244","PU245","PU246","PU247","AM","AM231","AM232",
"AM233","AM234","AM235","AM236","AM237","AM238","AM239","AM240",
"AM241","AM242","AM243","AM244","AM245","AM246","AM247","AM248",
"AM249","CM","CM233","CM234","CM235","CM236","CM237","CM238","CM239",
"CM240","CM241","CM242","CM243","CM244","CM245","CM246","CM247",
"CM248","CM249","CM250","CM251","CM252","BK","BK235","BK236","BK237",
"BK238","BK239","BK240","BK241","BK242","BK243","BK244","BK245",
"BK246","BK247","BK248","BK249","BK250","BK251","BK252","BK253",
"BK254","CF","CF237","CF238","CF239","CF240","CF241","CF242","CF243",
"CF244","CF245","CF246","CF247","CF248","CF249","CF250","CF251",
"CF252","CF253","CF254","CF255","CF256","ES","ES240","ES241","ES242",
"ES243","ES244","ES245","ES246","ES247","ES248","ES249","ES250",
"ES251","ES252","ES253","ES254","ES255","ES256","ES257","ES258","FM",
"FM242","FM243","FM244","FM245","FM246","FM247","FM248","FM249",
"FM250","FM251","FM252","FM253","FM254","FM255","FM256","FM257",
"FM258","FM259","FM260","MD","MD245","MD246","MD247","MD248","MD249",
"MD250","MD251","MD252","MD253","MD254","MD255","MD256","MD257",
"MD258","MD259","MD260","MD261","MD262","NO","NO248","NO249","NO250",
"NO251","NO252","NO253","NO254","NO255","NO256","NO257","NO258",
"NO259","NO260","NO261","NO262","NO263","NO264","LR","LR251","LR252",
"LR253","LR254","LR255","LR256","LR257","LR258","LR259","LR260",
"LR261","LR262","LR263","LR264","LR265","LR266","RF","RF253","RF254",
"RF255","RF256","RF257","RF258","RF259","RF260","RF261","RF262",
"RF263","RF264","RF265","RF266","RF267","RF268","DB","DB255","DB256",
"DB257","DB258","DB259","DB260","DB261","DB262","DB263","DB264",
"DB265","DB266","DB267","DB268","DB269","DB270","SG","SG258","SG259",
"SG260","SG261","SG262","SG263","SG264","SG265","SG266","SG267",
"SG268","SG269","SG270","SG271","SG272","SG273","BH","BH260","BH261",
"BH262","BH263","BH264","BH265","BH266","BH267","BH268","BH269",
"BH270","BH271","BH272","BH273","BH274","BH275","HS","HS263","HS264",
"HS265","HS266","HS267","HS268","HS269","HS270","HS271","HS272",
"HS273","HS274","HS275","HS276","HS277","MT","MT265","MT266","MT267",
"MT268","MT269","MT270","MT271","MT272","MT273","MT274","MT275",
"MT276","MT277","MT278","MT279","DS","DS267","DS268","DS269","DS270",
"DS271","DS272","DS273","DS274","DS275","DS276","DS277","DS278",
"DS279","DS280","DS281","RG","RG272","RG273","RG274","RG275","RG276",
"RG277","RG278","RG279","RG280","RG281","RG282","RG283","UUB",
"UUB277","UUB278","UUB279","UUB280","UUB281","UUB282","UUB283",
"UUB284","UUB285","UUT","UUT283","UUT284","UUT285","UUT286","UUT287",
"UUQ","UUQ285","UUQ286","UUQ287","UUQ288","UUQ289","UUP","UUP287",
"UUP288","UUP289","UUP290","UUP291","UUH","UUH289","UUH290","UUH291",
"UUH292","UUS","UUS291","UUS292","UUO","UUO293"};

static const double atomic_masses[] =
{1.00782503207, 1.00782503207, 2.01410177785, 2.01410177785, 3.01604927767,
3.01604927767, 4.027806424, 5.035311488, 6.044942594, 7.052749,
4.00260325415, 3.01602931914, 4.00260325415, 5.012223624, 6.018889124,
7.028020618, 8.033921897, 9.043950286, 10.052398837, 7.016004548, 3.030775,
4.027185558, 5.0125378, 6.015122794, 7.016004548, 8.022487362, 9.026789505,
10.035481259, 11.043797715, 12.053780, 9.012182201, 5.040790, 6.019726317,
7.016929828, 8.005305103, 9.012182201, 10.013533818, 11.021657749,
12.026920737, 13.035693007, 14.04289292, 15.053460, 16.061920, 11.009305406,
6.046810, 7.029917901, 8.024607233, 9.013328782, 10.012936992, 11.009305406,
12.014352104, 13.017780217, 14.025404009, 15.031103021, 16.039808829,
17.046989906, 18.056170, 19.063730, 12, 8.037675025, 9.031036689,
10.016853228, 11.011433613, 12, 13.00335483778, 14.0032419887, 15.010599256,
16.014701252, 17.022586116, 18.026759354, 19.034805018, 20.040319754,
21.049340, 22.057200, 14.00307400478, 10.041653674, 11.026090956,
12.018613197, 13.005738609, 14.00307400478, 15.00010889823, 16.006101658,
17.008450261, 18.014078959, 19.017028697, 20.023365807, 21.02710824,
22.034394934, 23.041220, 24.051040, 25.060660, 15.99491461956,
12.034404895, 13.024812213, 14.00859625, 15.003065617, 15.99491461956,
16.999131703, 17.999161001, 19.00358013, 20.004076742, 21.008655886,
22.009966947, 23.015687659, 24.020472917, 25.029460, 26.038340, 27.048260,
28.057810, 18.998403224, 14.035060, 15.018009103, 16.011465724,
17.002095237, 18.000937956, 18.998403224, 19.999981315, 20.999948951,
22.002998815, 23.003574631, 24.008115485, 25.012101747, 26.019615555,
27.026760086, 28.035670, 29.043260, 30.052500, 31.060429, 19.99244017542,
16.025761262, 17.017671504, 18.005708213, 19.001880248, 19.99244017542,
20.993846684, 21.991385113, 22.994466904, 23.993610779, 24.997736888,
26.000461206, 27.007589903, 28.012071575, 29.019385933, 30.024801045,
31.033110, 32.040020, 33.049380, 34.057028, 22.98976928087, 18.025969,
19.013877499, 20.007351328, 20.997655206, 21.994436425, 22.98976928087,
23.990962782, 24.989953968, 25.992633, 26.994076788, 27.998938, 29.002861,
30.008976, 31.013585452, 32.02046656, 33.026719756, 34.035170, 35.042493,
36.051480, 37.059340, 23.985041699, 19.03547, 20.018862545, 21.01171291,
21.999573843, 22.994123669, 23.985041699, 24.985836917, 25.982592929,
26.984340585, 27.983876825, 28.9886, 29.990434, 30.996546, 31.998975,
33.005254, 34.009456424, 35.017340, 36.023000, 37.031400, 38.037570,
39.046772, 40.053930, 26.981538627, 21.028040, 22.019520, 23.007267432,
23.999938865, 24.990428095, 25.986891692, 26.981538627, 27.981910306,
28.980445046, 29.982960256, 30.983946619, 31.988124489, 32.990843336,
33.996851837, 34.999860235, 36.006207204, 37.01067782, 38.017231021,
39.02297, 40.031450, 41.038330, 42.046890, 27.97692653246, 22.034530,
23.025520, 24.011545616, 25.004105574, 25.992329921, 26.986704905,
27.97692653246, 28.9764947, 29.973770171, 30.975363226999998,
31.974148082, 32.97800022, 33.978575524, 34.984583575, 35.986599477,
36.99293608, 37.995633601, 39.002070013, 40.005869121, 41.01456,
42.019790, 43.028660, 44.035260, 30.973761629, 24.034350, 25.020260,
26.011780, 26.999230236, 27.992314761, 28.981800606, 29.978313789,
30.973761629, 31.973907274, 32.971725543, 33.973636257, 34.973314117,
35.97825968, 36.979608946, 37.984156827, 38.986179475, 39.991296951,
40.994335435, 42.001007913, 43.00619, 44.012990, 45.019220, 46.027380,
31.972070999, 26.027880, 27.018833, 28.004372763, 28.996608049,
29.984903249, 30.979554728, 31.972070999, 32.971458759, 33.967866902,
34.969032161, 35.96708076, 36.971125567, 37.971163317, 38.975134306,
39.975451728, 40.979582149, 41.981022419, 42.98715479, 43.99021339,
44.996508112, 46.000750, 47.008590, 48.014170, 49.023619, 34.968852682,
28.028510, 29.014110, 30.004770, 30.992413086, 31.985689901, 32.977451887,
33.973762819, 34.968852682, 35.968306981, 36.965902591, 37.968010425,
38.968008164, 39.970415472, 40.970684525, 41.973254804, 42.974054403,
43.978281071, 44.980286886, 45.98421004, 46.988710, 47.994950, 49.000320,
50.007840, 51.014490, 39.96238312251, 30.021560, 31.012123, 31.997637984,
32.989925709, 33.980271244, 34.975257585, 35.967545105, 36.96677632,
37.962732394, 38.964313231, 39.96238312251, 40.964500611, 41.963045736,
42.965636056, 43.964924033, 44.968039956, 45.968094129, 46.972186792,
47.974540, 48.980520, 49.984430, 50.991630, 51.996780, 53.004940,
38.963706679, 32.021920, 33.007260, 33.998410, 34.988009692, 35.981292235,
36.973375889, 37.969081184, 38.963706679, 39.963998475, 40.961825762,
41.96240281, 42.96071554, 43.961556804, 44.960699493, 45.961976864,
46.961678473, 47.965513535, 48.967450928, 49.972783355, 50.976380,
51.982610, 52.987120, 53.994200, 54.999710, 39.962590983, 34.014120,
35.004940, 35.993087063, 36.985870269, 37.976318452, 38.970719725,
39.962590983, 40.962278062, 41.958618014, 42.958766628, 43.955481754,
44.956186566, 45.953692587, 46.954546006, 47.952534177, 48.955674148,
49.957518962, 50.961499214, 51.9651, 52.970050, 53.974350, 54.980550,
55.985570, 56.992356, 44.955911909, 36.014920, 37.003050, 37.994700,
38.984790002, 39.977967407, 40.969251125, 41.965516429, 42.961150658,
43.959402752, 44.955911909, 45.95517189, 46.952407508, 47.952231468,
48.950023975, 49.952187685, 50.953603368, 51.956675468, 52.959610,
53.963264561, 54.968243949, 55.972870, 56.977790, 57.983710, 58.989220,
59.995710, 47.947946281, 38.009770, 39.001610, 39.990498838, 40.983145,
41.973030902, 42.968522499, 43.959690069, 44.958125616, 45.952631555,
 46.951763088, 47.947946281, 48.947869982, 49.944791194, 50.946614955,
51.946897311, 52.949727171, 53.951052401, 54.955265056, 55.958199639,
56.963989137, 57.966970, 58.972930, 59.976760, 60.983200, 61.987490,
62.994420, 50.943959507, 40.011090, 40.999780, 41.991230, 42.980650,
43.97411, 44.965775808, 45.960200481, 46.95490894, 47.952253707,
48.948516101, 49.947158485, 50.943959507, 51.944775479, 52.944337979,
53.946439854, 54.947233701, 55.950530966, 56.952561432, 57.956834136,
58.960207407, 59.965026862, 60.968480, 61.973780, 62.977550, 63.983470,
64.987920, 51.940507472, 42.006430, 42.997710, 43.985549, 44.97964,
45.968358635, 46.962900046, 47.954031716, 48.951335721, 49.946044205,
50.944767431, 51.940507472, 52.940649386, 53.938880395, 54.940839672,
55.940653139, 56.943613013, 57.944353129, 58.948586367, 59.950076033,
60.954717204, 61.95661319, 62.961860, 63.964410, 64.970160, 65.973380,
66.979550, 54.938045141, 44.006870, 44.994510, 45.986720, 46.976100,
47.96852, 48.959618005, 49.95423823, 50.948210787, 51.945565464,
52.941290117, 53.940358854, 54.938045141, 55.93890491, 56.938285378,
57.939981549, 58.940440237, 59.942911246, 60.944652638, 61.94842822,
62.95023999, 63.95424909, 64.956336065, 65.961080, 66.964140, 67.969300,
68.972840, 55.934937475, 45.014578, 46.000810, 46.992890, 47.980504,
48.973610, 49.962988982, 50.956819538, 51.948113875, 52.945307942,
53.939610501, 54.938293357, 55.934937475, 56.935393969, 57.933275558,
58.934875464, 59.934071683, 60.936745281, 61.936767442, 62.940369091,
63.941201265, 64.94538027, 65.946780638, 66.950947244, 67.9537, 68.958780,
69.961460, 70.966720, 71.969620, 58.933195048, 47.011490, 48.001760,
48.989720, 49.981540, 50.970720, 51.963590, 52.954218896, 53.948459635,
54.941999029, 55.939839278, 56.936291373, 57.935752814, 58.933195048,
59.933817059, 60.932475763, 61.934050563, 62.933611611, 63.935809908,
64.93647846, 65.939762004, 66.940889529, 67.944873058, 68.94632, 69.951,
70.9529, 71.957810, 72.960240, 73.965380, 74.968330, 57.935342907,
48.019750, 49.009660, 49.995930, 50.987720, 51.975680, 52.968470,
53.957905495, 54.951330251, 55.942132022, 56.939793526, 57.935342907,
58.934346705, 59.930786372, 60.931056033, 61.928345115, 62.929669374,
63.927965959, 64.930084304, 65.929139334, 66.931569414, 67.931868789,
68.935610269, 69.9365, 70.940736283, 71.942092682, 72.946470, 73.948070,
74.952870, 75.955330, 76.960550, 77.963180, 62.929597474, 51.997180,
52.985550, 53.976710, 54.966050, 55.958560, 56.949211078, 57.944538499,
58.939498028, 59.93736503, 60.933457821, 61.932583745, 62.929597474,
63.929764183, 64.927789485, 65.928868813, 66.927730314, 67.929610889,
68.929429269, 69.932392343, 70.932676833, 71.935820307, 72.936675282,
73.939874862, 74.9419, 75.945275026, 76.947850, 77.951960, 78.954560,
79.960870, 63.929142222, 53.992950, 54.983980, 55.972380, 56.964788,
57.954591555, 58.949263764, 59.941827035, 60.939510635, 61.934329764,
62.933211566, 63.929142222, 64.929240984, 65.926033419, 66.927127345,
67.924844154, 68.926550281, 69.925319274, 70.927721599, 71.926857951,
72.929779104, 73.929458609, 74.932936741, 75.93329357, 76.936958967,
77.938440216, 78.942652, 79.944342348, 80.950480, 81.954420, 82.961030,
68.925573587, 55.994910, 56.982930, 57.974250, 58.963370, 59.957060,
60.949446287, 61.944175238, 62.939294196, 63.936838747, 64.932734754,
65.93158901, 66.928201703, 67.927980084, 68.925573587, 69.926021972,
70.924701349, 71.926366268, 72.925174682, 73.926945762, 74.926500246,
75.928827626, 76.9291543, 77.93160818, 78.93289326, 79.936515781,
80.937752355, 81.942990, 82.946980, 83.952650, 84.957000, 85.963120,
73.921177767, 57.991010, 58.981750, 59.970190, 60.963790, 61.954650,
62.949640, 63.941653, 64.939436406, 65.933843453, 66.93273407,
67.92809424, 68.927964533, 69.924247381, 70.924950954, 71.922075815,
72.923458945, 73.921177767, 74.922858948, 75.921402557, 76.923548591,
77.922852739, 78.925400995, 79.925372392, 80.928820467, 81.929549725,
82.934620, 83.937470, 84.943030, 85.946490, 86.952510, 87.956910,
88.963830, 74.921596478, 59.993130, 60.980620, 61.973200, 62.963690,
63.957572, 64.949564, 65.94471, 66.939186071, 67.936769069, 68.932273675,
69.930924826, 70.927112428, 71.926752283, 72.923824844, 73.923928692,
74.921596478, 75.922394021, 76.920647286, 77.921827281, 78.920947934,
79.922533816, 80.922132287, 81.924504067, 82.924980024, 83.929058,
84.932020, 85.936500, 86.939900, 87.944940, 88.949390, 89.955500,
90.960430, 91.966800, 79.916521271, 64.964660, 65.955210, 66.950090,
67.941798, 68.939557817, 69.933390644, 70.932241822, 71.927112352,
72.926765345, 73.922476436, 74.922523368, 75.919213597, 76.919914038,
77.91730909, 78.918499098, 79.916521271, 80.917992474, 81.916699401,
82.919118473, 83.918462354, 84.922245053, 85.924271579, 86.928521358,
87.931423998, 88.936450, 89.939960, 90.945960, 91.949920, 92.956290,
93.960490, 78.918337087, 66.964790, 67.958516, 68.950106, 69.944792,
70.93874, 71.936644572, 72.931691524, 73.929891034, 74.925776207,
75.924541469, 76.921379082, 77.921145706, 78.918337087, 79.918529296,
80.916290563, 81.916804119, 82.915180421, 83.916478974, 84.915608403,
85.918797577, 86.920711324, 87.924065926, 88.926385334, 89.930627737,
90.933968095, 91.939258714, 92.943050, 93.948680, 94.952870, 95.958530,
96.962800, 85.910610729, 68.965180, 69.955259, 70.949625738, 71.942092038,
72.939289195, 73.933084369, 74.930945746, 75.925910078, 76.92467,
77.920364783, 78.920082431, 79.916378965, 80.916592015, 81.9134836,
82.914136099, 83.911506687, 84.912527331, 85.910610729, 86.913354862,
87.914446969, 88.917630581, 89.919516555, 90.923445215, 91.92615621,
92.931274357, 93.934360, 94.939840, 95.943070, 96.948560, 97.951910,
98.957600, 99.961140, 84.911789737, 70.965320, 71.959080, 72.950561,
73.944264751, 74.93857, 75.935072226, 76.930408, 77.928141, 78.92398946,
79.92251925, 80.918995913, 81.918208598, 82.915109701, 83.914384821,
84.911789737, 85.911167419, 86.909180526, 87.911315588, 88.912278016,
89.914801694, 90.916536958, 91.9197289, 92.922041876, 93.926404946,
94.929302889, 95.934272637, 96.937351916, 97.941790668, 98.945379283,
99.949870, 100.953196445, 101.958870, 87.905612124, 72.965970,
73.956310, 74.949949568, 75.941766782, 76.937944782, 77.93218,
78.929708, 79.924521013, 80.923211846, 81.918401639, 82.917556701,
83.913425275, 84.912932803, 85.909260204, 86.908877124, 87.905612124,
88.907450675, 89.907737888, 90.910203095, 91.911037858, 92.914025634,
93.915361312, 94.919358766, 95.921696802, 96.926152923, 97.928452934,
98.933240926, 99.935351911, 100.940517888, 101.943018987, 102.948950,
103.952330, 104.958580, 88.905848295, 75.958450, 76.949645, 77.943610,
78.937351634, 79.93428, 80.929127468, 81.926792451, 82.922354243,
83.920388264, 84.916433039, 85.914885576, 86.91087573, 87.909501146,
88.905848295, 89.907151886, 90.907304791, 91.908949143, 92.909582713,
93.911595245, 94.912820621, 95.915891343, 96.918133995, 97.92220302,
98.924636204, 99.927756586, 100.93031385, 101.933555695, 102.936730,
103.941050, 104.944870, 105.949790, 106.954140, 107.959480,
89.904704416, 77.955230, 78.949160, 79.9404, 80.937210026, 81.931087,
82.928653801, 83.923250, 84.921471182, 85.916473591, 86.914816252,
87.910226904, 88.9088895, 89.904704416, 90.905645767, 91.905040847,
92.906476006, 93.906315192, 94.9080426, 95.908273386, 96.910953109,
97.912734892, 98.916512106, 99.917761889, 100.921140415, 101.922981285,
102.926599606, 103.928780, 104.933050, 105.935910, 106.940750,
107.943960, 108.949240, 109.952870, 92.906378058, 80.949030,
81.943130, 82.936705382, 83.933570, 84.927912447, 85.925038326,
86.920361108, 87.918332163, 88.913418245, 89.911264845,
90.906996243, 91.907193888, 92.906378058, 93.907283888, 94.906835792,
95.908100647, 96.908098556, 97.910328412, 98.911618375, 99.914181619,
100.915252025, 101.918037614, 102.919143842, 103.922464701,
104.923936545, 105.927970, 106.930310, 107.934840, 108.937630,
109.942440, 110.945650, 111.950830, 112.954700, 97.905408169, 82.948740,
83.940090, 84.936550, 85.930695904, 86.927326502, 87.921953241,
88.919480009, 89.913936896, 90.911750194, 91.906810991, 92.90681261,
93.905088269, 94.905842129, 95.904679477, 96.906021465, 97.905408169,
98.90771187, 99.907477336, 100.910347001, 101.91029736, 102.913207142,
103.913763625, 104.91697461, 105.918136802, 106.921692604, 107.923453,
108.927810, 109.929730, 110.934410, 111.936840, 112.941880, 113.944920,
114.950290, 98.906254747, 84.948830, 85.942880, 86.936530, 87.932678,
88.927167, 89.923556564, 90.918427639, 91.915260166, 92.910248984,
93.909657002, 94.907657084, 95.907871383, 96.906365358, 97.907215966,
98.906254747, 99.90765778, 100.907314659, 101.909215019, 102.909181351,
103.911447454, 104.911660566, 105.914357927, 106.915079572, 107.918461226,
108.919982665, 109.923820483, 110.92569283, 111.929146493, 112.931590,
113.935880, 114.938690, 115.943370, 116.946480, 117.951480, 101.904349312,
86.949180, 87.940260, 88.936110, 89.929890, 90.926292, 91.920120,
92.917052034, 93.911359711, 94.910412929, 95.907597835, 96.9075547,
97.905287132, 98.905939302, 99.904219476, 100.905582087, 101.904349312,
102.906323847, 103.905432701, 104.907752866, 105.907329433,
106.909905089, 107.910173465, 108.913203233, 109.914136041, 110.917696,
111.918965, 112.922487194, 113.924281, 114.928686173, 115.930810,
116.935580, 117.937820, 118.942840, 119.945310, 102.905504292,
88.948837, 89.942870, 90.936550, 91.931980, 92.925740, 93.921698,
94.91589874, 95.914460631, 96.911336797, 97.910708158, 98.908132104,
99.90812155, 100.906163625, 101.906843196, 102.905504292, 103.906655518,
104.905693821, 105.907287135, 106.906748423, 107.908728018, 108.908737289,
109.911136411, 110.911585913, 111.914394159, 112.915530627, 113.918806,
114.920334, 115.924062, 116.925980, 117.930070, 118.932110, 119.936410,
120.938720, 121.943210, 105.903485715, 90.949110, 91.940420, 92.935910,
93.928770, 94.924690, 95.918164359, 96.916479073, 97.912720902,
98.911767833, 99.908505886, 100.908289242, 101.905608544, 102.906087307,
103.904035834, 104.90508492, 105.903485715, 106.905133481, 107.903891701,
108.905950451, 109.905153254, 110.907670734, 111.907314058, 112.910152908,
113.910362638, 114.913683824, 115.914158662, 116.917841338, 117.9189843,
118.923110, 119.924691878, 120.928870, 121.930550, 122.934930, 123.936880,
106.90509682, 92.949780, 93.942780, 94.935480, 95.930680, 96.923972412,
97.921566201, 98.917597178, 99.916104255, 100.912802233, 101.911685,
102.90897272, 103.908629157, 104.906528661, 105.906668921, 106.90509682,
107.905955556, 108.904752292, 109.906107231, 110.905291157, 111.907004814,
112.906566579, 113.908803704, 114.908762698, 115.911359933, 116.911684562,
117.914582768, 118.915665059, 119.918787384, 120.919848046, 121.923530,
122.924900, 123.928640, 124.930430, 125.934500, 126.936770, 127.941170,
128.943690, 129.950448, 113.90335854, 94.949870, 95.939770, 96.934940,
97.927395546, 98.925010, 99.920289525, 100.918681538, 101.914462258,
102.913419246, 103.909849475, 104.909467905, 105.90645941, 106.906617928,
107.904183683, 108.904982293, 109.90300207, 110.904178107, 111.902757809,
112.904401662, 113.90335854, 114.905430969, 115.904755809, 116.907218618,
117.90691453, 118.909921597, 119.909850129, 120.912977363, 121.913332432,
122.917002999, 123.917647616, 124.92124637, 125.922353321, 126.926443864,
127.927762285, 128.932150, 129.933901937, 130.940670, 131.945550,
114.903878484, 96.949540, 97.942140, 98.934220, 99.931110851,
100.926340, 101.924090238, 102.919914188, 103.918296171, 104.91467354,
105.913465411, 106.9102951, 107.90969818, 108.907150507, 109.907165274,
110.905103278, 111.905532331, 112.904057761, 113.904913876,
114.903878484, 115.905259703, 116.904513564, 117.906354367, 118.90584535,
119.907959608, 120.907845822, 121.91027601, 122.910438276, 123.913175231,
124.913600588, 125.916463857, 126.917353091, 127.920172328, 128.92169698,
129.924970049, 130.926851767, 131.93299026, 132.937810, 133.944150,
134.949330, 119.902194676, 98.949330, 99.939044343, 100.936060,
101.930295324, 102.928100, 103.923143223, 104.921349437, 105.91688062,
106.915644329, 107.911925378, 108.911283214, 109.907842791, 110.90773446,
111.904818207, 112.905170577, 113.902778869, 114.903342397, 115.90174053,
116.902951656, 117.901603167, 118.90330763, 119.902194676, 120.90423548,
121.903439046, 122.905720838, 123.905273946, 124.907784125, 125.90765328,
126.910360024, 127.910536624, 128.913479, 129.913967295, 130.916999769,
131.917815713, 132.923829249, 133.928291765, 134.934730, 135.939340,
136.945990, 120.903815686, 102.939690, 103.936472, 104.931486348,
105.928791, 106.924150, 107.922160, 108.918132426, 109.916753, 110.913163,
111.912398009, 112.909371672, 113.909269, 114.906598, 115.906793629,
116.904835941, 117.905528731, 118.903942009, 119.905072427, 120.903815686,
121.905173651, 122.90421397, 123.905935743, 124.905253818, 125.90724748,
126.906923609, 127.909169001, 128.909148442, 129.911656324, 130.911982275,
131.914466896, 132.91525163, 133.920379744, 134.925165771, 135.930350,
136.935310, 137.940790, 138.945980, 129.906224399, 104.943640,
105.937504237, 106.935006, 107.929444597, 108.927415515, 109.922407316,
110.921110692, 111.917013672, 112.915891, 113.912089, 114.911902,
115.90846, 116.908644719, 117.905827581, 118.906403645, 119.904020222,
120.904936424, 121.903043898, 122.904270029, 123.902817896, 124.904430731,
125.903311696, 126.905226336, 127.904463056, 128.906598238, 129.906224399,
130.908523864, 131.90855316, 132.910955306, 133.911368737, 134.916448592,
135.920101246, 136.925322954, 137.929220, 138.934730, 139.938850,
140.944650, 141.949080, 126.904472681, 107.943475, 108.938149417,
109.935242, 110.930276, 111.927970, 112.923640583, 113.921850, 114.918048,
115.916808633, 116.91365, 117.913074, 118.910074, 119.910048173,
120.907366811, 121.907589284, 122.905588965, 123.906209852, 124.904630164,
125.905624153, 126.904472681, 127.905809443, 128.904987722, 129.906674247,
130.906124609, 131.907997381, 132.907796939, 133.909744465, 134.910048121,
135.914653993, 136.91787084, 137.922349591, 138.926099478, 139.931000,
140.935030, 141.940180, 142.944560, 143.949990, 131.904153457, 109.944278068,
110.941602, 111.935623112, 112.933341174, 113.927980306, 114.92629392,
115.921581087, 116.920358735, 117.916178655, 118.915410688, 119.911784244,
120.911461829, 121.908367632, 122.90848191, 123.905893003, 124.906395464,
125.904273634, 126.905183723, 127.903531275, 128.904779435, 129.903508007,
130.905082362, 131.904153457, 132.905910722, 133.905394464, 134.907227495,
135.907218794, 136.911562125, 137.913954475, 138.918792936, 139.921640943,
140.926648049, 141.92970959, 142.935110, 143.938510, 144.944070, 145.947750,
146.953560, 132.905451932, 111.950301, 112.944493274, 113.941450, 114.935910,
115.933367, 116.928670701, 117.926559494, 118.922377304, 119.920677253,
120.917229209, 121.916113434, 122.912996036, 123.912257798, 124.90972827,
125.909451977, 126.907417525, 127.907748866, 128.906064426, 129.906708552,
130.905463926, 131.90643426, 132.905451932, 133.906718475, 134.905977008,
135.907311576, 136.907089473, 137.911016704, 138.913363999, 139.917282354,
140.920045752, 141.924298927, 142.92735175, 143.932076914, 144.93552617,
145.940289423, 146.944155008, 147.949218153, 148.952930, 149.958170,
150.962190, 137.905247237, 113.950675405, 114.947370, 115.941380,
116.938499, 117.933040, 118.930659661, 119.926044974, 120.924054499,
121.919904, 122.918781036, 123.915093603, 124.914472912, 125.911250177,
126.911093797, 127.908317698, 128.908679439, 129.906320811, 130.906941118,
131.905061288, 132.90600749, 133.904508383, 134.905688591, 135.904575945,
136.905827384, 137.905247237, 138.908841341, 139.910604505, 140.914411009,
141.91645341, 142.920626719, 143.922952853, 144.927627032, 145.930219572,
146.934945, 147.937720047, 148.942580, 149.945680, 150.950810, 151.954270,
152.959610, 138.906353267, 116.950068, 117.946730, 118.940990, 119.938070,
120.933010, 121.930710, 122.926240, 123.924574275, 124.920816034,
125.919512667, 126.916375448, 127.915585177, 128.912692815, 129.912368724,
130.91007, 131.910101145, 132.908218, 133.908514011, 134.906976844,
135.907635536, 136.906493598, 137.90711193, 138.906353267, 139.909477645,
140.910962152, 141.91407913, 142.91606272, 143.919599647, 144.921645401,
145.92579346, 146.928235284, 147.932228868, 148.934734, 149.938770,
150.941720, 151.946250, 152.949620, 153.954500, 154.958350, 139.905438706,
118.952760, 119.946640, 120.943420, 121.937910, 122.935400, 123.930410,
124.928440, 125.923971, 126.922731, 127.918911, 128.918102, 129.914736,
130.914422, 131.911460487, 132.91151502, 133.908924821, 134.909151396,
135.907172422, 136.907805577, 137.905991321, 138.906652651, 139.905438706,
140.90827627, 141.909244205, 142.91238591, 143.913647336, 144.917233135,
145.918759009, 146.922673954, 147.92443241, 148.928399883, 149.930408931,
150.933976196, 151.936540, 152.940580, 153.943420, 154.948040, 155.951260,
156.956340, 140.907652769, 120.955364, 121.951810, 122.945960, 123.942960,
124.937830, 125.935310, 126.930830, 127.928791, 128.925095, 129.92359,
130.920259, 131.919255, 132.916330532, 133.915711737, 134.913111745,
135.912691611, 136.910705455, 137.910754636, 138.908938399, 139.909075874,
140.907652769, 141.910044806, 142.910816926, 143.913305245, 144.9145117,
145.917644336, 146.918995992, 147.922135026, 148.923717651, 149.926672997,
150.928318618, 151.931499225, 152.933838905, 153.937518153, 154.940120,
155.944270, 156.947430, 157.951980, 158.955500, 141.907723297, 123.952230,
124.948880, 125.943220, 126.940500, 127.935390, 128.933188, 129.928506,
130.927247, 131.923321237, 132.922348, 133.918790181, 134.91818116,
135.914976035, 136.914567137, 137.911949961, 138.911978288, 139.909552,
140.909609854, 141.907723297, 142.90981429, 143.910087274, 144.912573636,
145.913116939, 146.916100441, 147.916893288, 148.920148842, 149.920890888,
150.923828929, 151.924682219, 152.927698232, 153.929477307, 154.932932,
155.935018114, 156.939030, 157.941600, 158.946090, 159.949090, 160.953880,
144.912749023, 125.957520, 126.951630, 127.948420, 128.943160, 129.940450,
130.935870, 131.933750, 132.929782, 133.928353, 134.924876, 135.923565829,
136.920479493, 137.919548281, 138.916804082, 139.916041789, 140.913555054,
141.912874471, 142.910932616, 143.912590843, 144.912749023, 145.914696305,
146.915138545, 147.917474618, 148.918334155, 149.920983561, 150.921206973,
151.923496795, 152.924116889, 153.926463943, 154.928101267, 155.931056736,
156.933039369, 157.936561407, 158.938970, 159.942990, 160.945860,
161.950290, 162.953680, 151.919732425, 127.958080, 128.954640, 129.948920,
130.946110, 131.940690, 132.938670, 133.933970, 134.93252, 135.928275527,
136.926971746, 137.923243961, 138.922296605, 139.918994687, 140.918476488,
141.915197641, 142.914628338, 143.911999478, 144.913410353, 145.9130409,
146.914897923, 147.914822674, 148.917184735, 149.917275539, 150.919932409,
151.919732425, 152.922097356, 153.922209273, 154.924640161, 155.925527887,
156.928358717, 157.929991317, 158.933211271, 159.935140, 160.938830,
161.941220, 162.945360, 163.948280, 164.952980, 152.921230339, 129.963569,
130.957753, 131.954370, 132.949240, 133.946510, 134.941820, 135.939600,
136.935570, 137.933709, 138.92979228, 139.928087607, 140.92493072,
141.923434945, 142.920297509, 143.918816823, 144.916265237, 145.917205817,
146.916746111, 147.918085895, 148.917931238, 149.919701819, 150.919850161,
151.921744534, 152.921230339, 153.922979237, 154.92289326, 155.924752249,
156.925423647, 157.927845302, 158.929088861, 159.931971, 160.933680,
161.937040, 162.939210, 163.942990, 164.945720, 165.949970, 166.953210,
157.924103912, 133.955370, 134.952570, 135.947340, 136.945020, 137.940120,
138.938240, 139.933674, 140.932126, 141.928116, 142.92674951, 143.922963,
144.921709252, 145.918310608, 146.91909442, 147.918114524, 148.919340915,
149.918658876, 150.920348482, 151.919790996, 152.921749543, 153.920865598,
154.922622022, 155.922122743, 156.923960135, 157.924103912, 158.926388658,
159.927054146, 160.929669211, 161.930984751, 162.933990, 163.935860,
164.939380, 165.941600, 166.945570, 167.948360, 168.952870, 158.925346757,
135.961380, 136.955980, 137.953160, 138.948290, 139.945805049, 140.941448,
141.938744, 142.935121, 143.933045, 144.929274, 145.927246584, 146.924044585,
147.924271701, 148.923245909, 149.923659686, 150.923102543, 151.924074438,
152.923434588, 153.924678019, 154.923505236, 155.924747213, 156.924024604,
157.925413137, 158.925346757, 159.927167606, 160.927569919, 161.929488234,
162.930647536, 163.933350838, 164.934880, 165.937991959, 166.940050,
167.943640, 168.946220, 169.950250, 170.953300, 163.929174751, 137.962490,
138.959540, 139.954010, 140.951350, 141.946366, 142.943830, 143.939254,
144.937425, 145.932845369, 146.9310915, 147.927149831, 148.927304787,
149.925585184, 150.926184601, 151.9247183, 152.92576467, 153.924424457,
154.925753775, 155.92428311, 156.925466095, 157.924409487, 158.925739214,
159.925197517, 160.926933364, 161.926798447, 162.928731159, 163.929174751,
164.931703333, 165.932806741, 166.935655462, 167.937128769, 168.940307614,
169.942390, 170.946200, 171.948760, 172.953000, 164.93032207, 139.968539,
140.963098, 141.959770, 142.954610, 143.951480, 144.947200, 145.944640,
146.940056, 147.937718, 148.933774771, 149.933496182, 150.931688142,
151.931713714, 152.930198789, 153.930601579, 154.929103491, 155.929839,
156.928256188, 157.928941007, 158.927711959, 159.928729478, 160.927854776,
161.929095504, 162.928733903, 163.930233507, 164.93032207, 165.932284162,
166.933132633, 167.935515708, 168.936872273, 169.939618929, 170.94146515,
171.944820, 172.947290, 173.951150, 174.954050, 165.930293061, 142.966340,
143.960380, 144.957390, 145.952000, 146.949490, 147.944550, 148.942306,
149.937913839, 150.937448903, 151.935050389, 152.935063492, 153.932783081,
154.933208949, 155.931064698, 156.931916, 157.929893474, 158.930684066,
159.929083292, 160.929995309, 161.928778264, 162.930032749, 163.929200229,
164.930726003, 165.930293061, 166.932048159, 167.932370224, 168.934590364,
169.935464312, 170.938029808, 171.939356113, 172.942400, 173.944230,
174.947770, 175.950080, 176.954050, 168.93421325, 144.970073, 145.966425,
146.960961, 147.957840, 148.952720, 149.949960, 150.94548349, 151.944422,
152.942012112, 153.941567808, 154.939199459, 155.938979933, 156.936973,
157.936979525, 158.934975, 159.935262801, 160.933549, 161.933994682,
162.932651124, 163.93356, 164.932435492, 165.933554131, 166.932851622,
167.934172776, 168.93421325, 169.935801397, 170.93642944, 171.938400044,
172.939603607, 173.942168605, 174.943836853, 175.946994685, 176.949040,
177.952640, 178.955340, 173.938862089, 147.967420, 148.964040, 149.958420,
150.955400769, 151.950288919, 152.949480, 153.946393928, 154.945782332,
155.942818215, 156.942627848, 157.939865617, 158.940050099, 159.937552344,
160.937901678, 161.93576821, 162.936334305, 163.934489416, 164.935279,
165.933882042, 166.934949605, 167.933896895, 168.935189802, 169.934761837,
170.936325799, 171.936381469, 172.938210787, 173.938862089, 174.94127645,
175.942571683, 176.945260822, 177.94664668, 178.950170, 179.952330,
180.956150, 174.940771819, 149.973228, 150.967577, 151.964120,
152.958767331, 153.957522, 154.954316216, 155.953032523, 156.9500983,
157.949313283, 158.946628776, 159.946033, 160.943572, 161.943277288,
162.941179, 163.941339, 164.939406724, 165.939859, 166.93827,
167.938739111, 168.937651439, 169.938474968, 170.937913136, 171.939085669,
172.938930602, 173.94033748, 174.940771819, 175.94268631, 176.943758055,
177.945954559, 178.947327443, 179.94988116, 180.951970, 181.955040,
182.957570, 183.960910, 179.946549953, 152.970690, 153.964860, 154.963390,
155.959364025, 156.958396, 157.954799366, 158.95399487, 159.950684379,
160.950274844, 161.947210498, 162.947089, 163.944367284, 164.944567,
165.94218, 166.9426, 167.940568, 168.941259, 169.939609, 170.940492,
171.939448301, 172.940513, 173.940046178, 174.941509181, 175.941408631,
176.943220651, 177.943698766, 178.945816145, 179.946549953, 180.949101246,
181.950554096, 182.953530439, 183.955446515, 184.958820, 185.960890,
186.964590, 187.966850, 180.947995763, 154.974592, 155.972303,
156.968192445, 157.966699, 158.963018173, 159.961486056, 160.958417,
161.957291859, 162.954330271, 163.953534, 164.950772514, 165.950512,
166.948093, 167.948047, 168.946011, 169.946175, 170.944476, 171.944895,
172.94375, 173.944454, 174.943737, 175.944857, 176.944472403,
177.945778221, 178.945929535, 179.947464831, 180.947995763, 181.950151849,
182.951372616, 183.954007966, 184.955559375, 185.958552023, 186.960530,
187.963700, 188.965830, 189.969230, 183.950931188, 157.974562, 158.972918,
159.968478805, 160.967357, 161.963497417, 162.962523542, 163.958954382,
164.958279949, 165.955027253, 166.954816014, 167.951808394, 168.95177879,
169.949228482, 170.949451, 171.947292, 172.947689, 173.946079, 174.946717,
175.945634, 176.946643, 177.945876236, 178.947070447, 179.946704459,
180.948197248, 181.948204156, 182.950222951, 183.950931188, 184.953419264,
185.954364127, 186.957160466, 187.958489105, 188.961912868, 189.963181378,
190.966600, 191.968170, 186.955753109, 159.982115, 160.977589119,
161.976002, 162.972080535, 163.970323, 164.967088557, 165.965808,
166.962601, 167.961572608, 168.958791096, 169.958220071, 170.955716,
171.955422961, 172.953243, 173.953115, 174.951381, 175.951623, 176.950328,
177.950989, 178.949987641, 179.950789084, 180.950067916, 181.95121008,
182.950819841, 183.952520756, 184.952954982, 185.954986084, 186.955753109,
187.958114438, 188.959229007, 189.961817977, 190.963125242, 191.965960,
192.967470, 193.970420, 191.96148069, 161.984431, 162.982690,
163.978035649, 164.976762, 165.972690753, 166.971547969, 167.967803678,
168.96701927, 169.963577028, 170.963184819, 171.960023303, 172.959808409,
173.957062202, 174.956945835, 175.954806, 176.954965324, 177.953251241,
178.953816017, 179.952378803, 180.953244, 181.952110186, 182.953126102,
183.952489071, 184.954042265, 185.953838158, 186.955750458, 187.955838228,
188.95814747, 189.958447048, 190.960929718, 191.96148069, 192.964151563,
193.965182083, 194.968126661, 195.969639333, 192.96292643, 163.992201,
164.987520, 165.985824, 166.981665156, 167.979881, 168.976294942, 169.974965,
170.971626042, 171.970456, 172.967501739, 173.966861045, 174.964112895,
175.963648688, 176.9613015, 177.961082, 178.959122266, 179.959229446,
180.957625297, 181.958076296, 182.956846458, 183.957476, 184.956698,
185.957946104, 186.957363361, 187.958853121, 188.958718935, 189.960545968,
190.960594046, 191.962605012, 192.96292643, 193.965078378, 194.965979573,
195.968396542, 196.969653285, 197.972280, 198.973804583, 194.964791134,
165.994855, 166.992979, 167.988150742, 168.986715, 169.982495289,
170.981244542, 171.977347128, 172.976444754, 173.972818767, 174.972420552,
175.968944622, 176.968469481, 177.965648724, 178.965363404, 179.963031477,
180.963097285, 181.961170656, 182.961596703, 183.959922251, 184.960619,
185.959350813, 186.960587, 187.959395391, 188.960833686, 189.959931655,
190.961676661, 191.961038005, 192.962987401, 193.962680253, 194.964791134,
195.964951521, 196.967340182, 197.96789279, 198.970593094, 199.971440677,
200.974512868, 201.975740, 196.966568662, 168.998080, 169.996122,
170.991878881, 171.990035, 172.98623738, 173.984761, 174.981274107,
175.980099, 176.976864908, 177.97603192, 178.973212812, 179.972521124,
180.970079048, 181.969617874, 182.967593034, 183.967451524, 184.965789411,
185.965952703, 186.964567541, 187.965323661, 188.963948286, 189.964700339,
190.963704225, 191.964812953, 192.964149715, 193.96536525, 194.96503464,
195.966569813, 196.966568662, 197.968242303, 198.968765193, 199.970725647,
200.97165724, 201.973805838, 202.975154542, 203.977724, 204.979870,
201.970643011, 171.003760, 171.998832686, 172.997242, 173.992863695,
174.99142327, 175.98735458, 176.986279158, 177.982483143, 178.981833861,
179.978266394, 180.977819311, 181.974689964, 182.974449841, 183.971713051,
184.971899086, 185.96936179, 186.969814236, 187.967577049, 188.968190034,
189.966322449, 190.967157105, 191.965634327, 192.966665421, 193.965439409,
194.966720113, 195.965832649, 196.967212908, 197.966769032, 198.968279932,
199.968326004, 200.970302268, 201.970643011, 202.972872484, 203.973493933,
204.976073386, 205.977514066, 206.982588545, 207.985940, 208.991040,
209.994510, 204.974427541, 176.000590, 176.996427286, 177.994897,
178.991089082, 179.989906, 180.986257447, 181.985667104, 182.982192802,
183.981873122, 184.978791305, 185.978325, 186.975905897, 187.976009782,
188.973588428, 189.973877149, 190.971786154, 191.972225, 192.970672,
193.9712, 194.969774335, 195.970481151, 196.969574511, 197.970483495,
198.969877, 199.970962672, 200.970818891, 201.972105808, 202.97234422,
203.973863522, 204.974427541, 205.97611032, 206.977419429, 207.9820187,
208.985358952, 209.990073689, 210.993477, 211.998228, 207.976652071,
178.003830191, 179.002150, 179.997918173, 180.996623958, 181.992671842,
182.991874629, 183.988142339, 184.987609944, 185.984238945, 186.98391837,
187.980874338, 188.980807, 189.978081517, 190.978265, 191.975785171,
192.976173234, 193.97401207, 194.97454205, 195.972774109, 196.973431124,
197.972033959, 198.97291665, 199.971826675, 200.972884511, 201.972159133,
202.973390521, 203.973043589, 204.974481755, 205.974465278, 206.975896887,
207.976652071, 208.98109012, 209.984188527, 210.988736964, 211.991897543,
212.996581499, 213.999805408, 215.004807, 208.980398734, 184.001124,
184.997625, 185.996597625, 186.993157835, 187.992265154, 188.989199012,
189.988295129, 190.985786119, 191.985457954, 192.982959771, 193.98283396,
194.980650737, 195.980666509, 196.978864454, 197.979206, 198.977671961,
199.978131829, 200.977009036, 201.977742324, 202.976876001, 203.977812736,
204.977389366, 205.97849913, 206.978470679, 207.979742196, 208.980398734,
209.984120371, 210.98726946, 211.991285724, 212.994384666, 213.998711539,
215.001769776, 216.006305943, 217.009470, 218.014316, 208.982430435,
187.999422048, 188.998480562, 189.995101185, 190.994574485, 191.991335149,
192.991025275, 193.988185606, 194.988110728, 195.98553458, 196.98565963,
197.983388616, 198.983666063, 199.981798604, 200.982259764, 201.980757541,
202.981420103, 203.980318121, 204.981203322, 205.980481099, 206.981593173,
207.981245702, 208.982430435, 209.982873673, 210.986653154, 211.988867969,
212.99285728, 213.99520135, 214.999419988, 216.001915035, 217.006334796,
218.008973037, 219.013744, 220.016602, 210.987496271, 192.999843112,
193.998725085, 194.996268098, 195.995788077, 196.993189215, 197.992837202,
198.990532254, 199.990351264, 200.988416999, 201.988630236, 202.986941984,
203.987251326, 204.986074483, 205.986667036, 206.985783502, 207.986589977,
208.986173143, 209.98714771, 210.987496271, 211.990744771, 212.992936646,
213.996371733, 214.99865257, 216.002423257, 217.004718822, 218.008694336,
219.011161691, 220.015407682, 221.018050, 222.022330, 223.025190,
222.017577738, 195.005437696, 196.002115223, 197.001584351, 197.998678663,
198.998370297, 199.9956993, 200.995628335, 201.993263492, 202.993386687,
203.99142874, 204.991718799, 205.990214104, 206.990734225, 207.98964247,
208.990414742, 209.989696216, 210.990600523, 211.990703529, 212.993882668,
213.995362554, 214.998745483, 216.00027437, 217.003927675, 218.005601256,
219.009480204, 220.011393981, 221.015536782, 222.017577738, 223.021790,
224.024090, 225.028440, 226.030890, 227.035407, 228.037986, 222.01755173,
199.007258147, 200.00657249, 201.003860867, 202.003372847, 203.000924647,
204.000653204, 204.99859396, 205.998666066, 206.996949414, 207.997138783,
208.995953555, 209.996407738, 210.995536544, 211.996202244, 212.996189081,
213.998971145, 215.000341497, 216.00319799, 217.004631951, 218.007578322,
219.009252149, 220.012327405, 221.014254762, 222.01755173, 223.019735857,
224.023249951, 225.025565414, 226.029386231, 227.031835938, 228.035729,
229.038450228, 230.042510, 231.045440, 232.049772, 228.031070292,
202.009890686, 203.009271619, 204.006499668, 205.00626857, 206.00382727,
207.003798105, 208.00183994, 209.001991373, 210.000494978, 211.000897987,
211.999794499, 213.000383959, 214.000107894, 215.002719834, 216.003533035,
217.006320327, 218.00714023, 219.010085078, 220.011028384, 221.013917338,
222.01537453, 223.018502171, 224.020211821, 225.023611564, 226.025409823,
227.029177842, 228.031070292, 229.034957577, 230.037056394, 231.041220,
232.043638, 233.048060, 234.050704, 227.027752127, 206.01450498,
207.011949748, 208.011551551, 209.009494863, 210.009435986, 211.007734835,
212.007813822, 213.006607643, 214.006901798, 215.006453625, 216.008720075,
217.009346914, 218.011641453, 219.012420389, 220.014762979, 221.015591248,
222.017843851, 223.019137468, 224.021722866, 225.023229585, 226.026098089,
227.027752127, 228.031021112, 229.033015243, 230.036294178, 231.038558786,
232.042027438, 233.044550, 234.048420, 235.051232, 236.055296,
232.038055325, 209.017715682, 210.015075342, 211.014928413, 212.012980288,
213.01301014, 214.01149977, 215.01173033, 216.011062115, 217.013114328,
218.013284499, 219.015536895, 220.015747762, 221.018183674, 222.018468121,
223.020811448, 224.021466895, 225.023951021, 226.024903069, 227.02770407,
228.028741127, 229.03176243, 230.033133843, 231.036304343, 232.038055325,
233.041581843, 234.04360123, 235.047510074, 236.049870, 237.053894,
238.056496, 231.03588399, 212.023204138, 213.02110934, 214.020918417,
215.019185865, 216.019109564, 217.018323986, 218.020041889, 219.019883143,
220.021875303, 221.021877983, 222.023742, 223.023962273, 224.025625738,
225.026130678, 226.027947753, 227.028805072, 228.031051376, 229.032096793,
230.034540754, 231.03588399, 232.038591592, 233.040247277, 234.043308058,
235.045443615, 236.048681284, 237.051145659, 238.05450271, 239.057260,
240.060980, 238.050788247, 217.024368791, 218.023535671, 219.02491916,
220.024723, 221.026399, 222.026086, 223.0277386, 224.027604778,
225.029390717, 226.029338702, 227.031156367, 228.031374006, 229.033505939,
230.033939784, 231.036293704, 232.037156152, 233.039635207, 234.040952088,
235.043929918, 236.045568006, 237.048730184, 238.050788247, 239.054293299,
240.056591988, 241.060330, 242.062931, 237.048173444, 225.033913933,
226.035145, 227.034956789, 228.036180, 229.036263808, 230.037827597,
231.038245085, 232.040108, 233.040740546, 234.042895038, 235.044063267,
236.0465696, 237.048173444, 238.050946405, 239.052939025, 240.056162182,
241.058252431, 242.06164118, 243.064279, 244.067850, 242.058742611,
228.038742328, 229.040150212, 230.039649886, 231.041101107, 232.041187097,
233.042997375, 234.043317076, 235.04528605, 236.046057964, 237.048409658,
238.049559894, 239.052163381, 240.053813545, 241.056851456, 242.058742611,
243.062003092, 244.064203907, 245.067747154, 246.070204627, 247.074070,
243.06138108, 231.045560, 232.046590, 233.046348, 234.047809, 235.047946,
236.049579, 237.049996, 238.051984324, 239.053024479, 240.055300179,
241.056829144, 242.059549159, 243.06138108, 244.064284847, 245.066452114,
246.069774619, 247.072093, 248.075752, 249.078480, 247.07035354,
233.050771232, 234.050159841, 235.051434, 236.051413, 237.052901,
238.053028697, 239.054957, 240.055529539, 241.057653001, 242.058835824,
243.061389114, 244.062752578, 245.065491249, 246.067223662, 247.07035354,
248.072348508, 249.075953413, 250.078356959, 251.082284605, 252.084870,
247.07030708, 235.056580, 236.057330, 237.057003, 238.058281, 239.058279,
240.059759, 241.060230, 242.061981, 243.063007572, 244.065180774,
245.066361616, 246.068672947, 247.07030708, 248.073086, 249.074986657,
250.07831652, 251.080760172, 252.084310, 253.086880, 254.090600,
251.079586788, 237.062070, 238.061410, 239.062422, 240.062302, 241.063726,
242.063701552, 243.065427, 244.066000689, 245.068048612, 246.068805309,
247.071000589, 248.072184861, 249.074853537, 250.076406066, 251.079586788,
252.081625846, 253.085133145, 254.087322909, 255.091046, 256.093440,
252.082978512, 240.068920, 241.068538, 242.069745, 243.069548, 244.070883,
245.071324, 246.072896, 247.073656, 248.075471, 249.076411, 250.078612,
251.079992142, 252.082978512, 253.084824697, 254.088022021, 255.090273122,
256.093598, 257.095979, 258.099520, 257.095104724, 242.073430, 243.074353,
244.074084, 245.075385, 246.075299023, 247.076847, 248.077194714,
249.079034, 250.079521264, 251.081575017, 252.082466855, 253.085185236,
254.08685422, 255.089962202, 256.091773117, 257.095104724, 258.097076,
259.100595, 260.102678, 258.098431319, 245.080829, 246.081886, 247.081635,
248.082823, 249.083013, 250.084420, 251.084839, 252.086560, 253.087280,
254.089656, 255.091082705, 256.094059025, 257.095541368, 258.098431319,
259.100509, 260.103652, 261.105721, 262.108865, 255.093241131, 248.086596,
249.087833, 250.087510, 251.089012, 252.088976521, 253.090678,
254.090955253, 255.093241131, 256.094282666, 257.09687719, 258.098207,
259.101031, 260.102643, 261.105749, 262.107301, 263.110552, 264.112345,
260.105504, 251.094360, 252.095371, 253.095210, 254.096454, 255.096681,
256.098629, 257.099555, 258.101814, 259.102901, 260.105504, 261.106883,
262.109634, 263.111293, 264.114038, 265.115839, 266.119305, 263.112547,
253.100689, 254.100184, 255.101340, 256.101166194, 257.102990,
258.103489, 259.105637, 260.106440, 261.108766556, 262.109925, 263.112547,
264.113985, 265.116704, 266.117956, 267.121529, 268.123644, 255.107398,
255.107398, 256.108127, 257.107722, 258.109231, 259.109610, 260.111300,
261.112056, 262.114084, 263.114988, 264.117404, 265.118601, 266.121029,
267.122377, 268.125445, 269.127460, 270.130712, 259.114500, 258.113168,
259.114500, 260.114422071, 261.116117, 262.116398, 263.118322, 264.118931,
265.121114693, 266.122065, 267.124425, 268.125606, 269.128755, 270.130329,
271.133472, 272.135158, 273.138220, 262.122892, 260.121970, 261.121664,
262.122892, 263.123035, 264.124604, 265.125147, 266.126942, 267.127650,
268.129755, 269.130694, 270.133616, 271.135179, 272.138032, 273.139618,
274.142440, 275.144250, 263.128558, 263.128558, 264.128394885, 265.130085,
266.130097, 267.131789, 268.132162, 269.134056, 270.134650, 271.137657,
272.139052, 273.141986, 274.143131, 275.145952, 276.147208, 277.149841,
265.136151, 265.136151, 266.137299, 267.137307, 268.138728, 269.139055,
270.140657, 271.141139, 272.143738, 273.144913, 274.147492, 275.148647,
276.151156, 277.152420, 278.154812, 279.156193, 281.162061, 267.144341,
268.143795, 269.145124, 270.144720, 271.146062, 272.146317, 273.148863,
274.149492, 275.152176, 276.153034, 277.155647, 278.156469, 279.158861,
280.159795, 281.162061, 272.153615, 272.153615, 273.153682, 274.155713,
275.156142, 276.158493, 277.159519, 278.161604, 279.162468, 280.164473,
281.165372, 282.167486, 283.168415, 283.171792, 277.163943, 278.164312,
279.166546, 280.167039, 281.169286, 282.169765, 283.171792, 284.172384,
285.174105, 283.176451, 283.176451, 284.178080, 285.178732, 286.180481,
287.181045, 285.183698, 285.183698, 286.183855, 287.185599, 288.185689,
289.187279, 287.191186, 287.191186, 288.192492, 289.192715, 290.194141,
291.194384, 292.199786, 289.198862, 290.198590, 291.200011, 292.199786,
291.206564, 291.206564, 292.207549, 293.214670, 293.214670};

#endif /* header guard */
